/**********************************************************************************
 Copyright (C) 2009
 by 52 North Initiative for Geospatial Open Source Software GmbH

 Contact: Andreas Wytzisk 
 52 North Initiative for Geospatial Open Source Software GmbH
 Martin-Luther-King-Weg 24
 48155 Muenster, Germany
 info@52north.org

 This program is free software; you can redistribute and/or modify it under the
 terms of the GNU General Public License version 2 as published by the Free
 Software Foundation.

 This program is distributed WITHOUT ANY WARRANTY; even without the implied
 WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License along with this 
 program (see gnu-gplv2.txt). If not, write to the Free Software Foundation, Inc., 
 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or visit the Free Software
 Foundation web page, http://www.fsf.org.
 
 Created on: 18.11.2009
 *********************************************************************************/

package org.n52.spsMockup;

import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.xmlbeans.XmlError;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.sun.org.apache.xml.internal.serialize.OutputFormat;
import com.sun.org.apache.xml.internal.serialize.XMLSerializer;

/**
 * 
 * Class for helper methods dealing with XML, especially deeper inspection of XML-Documents respectively beans
 * (i.e. instances of (subclasses of) {@link XmlObject}).
 * 
 * @author Daniel Nüst (daniel.nuest@uni-muenster.de)
 * 
 */
public class XmlTools {

    private static final int DEFAULT_INDENT = 4;

    public static final XmlOptions PRETTY_PRINT_OPTIONS = new XmlOptions().setSavePrettyPrint().setSavePrettyPrintIndent(DEFAULT_INDENT);

    private static final int MAX_INSPECT_STRING_LENGTH = 142;

    public static final String OM_1_0_NAMESPACE_URI = "http://www.opengis.net/om/1.0";

    public static final String OM_NAMESPACE_PREFIX = "om";

    public static final String SOS_1_0_NAMESPACE_URI = "http://www.opengis.net/sos/1.0";

    public static final String SOS_NAMESPACE_PREFIX = "sos";

    public static final String OGC_NAMESPACE_URI = "http://www.opengis.net/ogc";

    public static final String OGC_NAMESPACE_PREFIX = "ogc";

    public static final String SWE_1_0_1_NAMESPACE_URI = "http://www.opengis.net/swe/1.0.1";

    public static final String SWE_NAMESPACE_PREFIX = "swe";

    public static final String SA_1_0_NAMESPACE_URI = "http://www.opengis.net/sampling/1.0";

    public static final String SA_NAMESPACE_PREFIX = "sa";

    public static final String GML_NAMESPACE_URI = "http://www.opengis.net/gml";

    public static final String OWS_NAMESPACE_PREFIX = "ows";

    public static final String OWS_1_1_NAMESPACE_URI = "http://www.opengis.net/ows/1.1";

    public static final String GML_NAMESPACE_PREFIX = "gml";

    public static final String XSI_NAMESPACE_PREFIX = "xsi";

    public static final String SML_NAMESPACE_PREFIX = "sml";

    public static final String SML_1_0_1_NAMESPACE_URI = "http://www.opengis.net/sensorML/1.0.1";

    public static final String SML_SCHEMA_LOCATION = "http://schemas.opengis.net/sensorML/1.0.1/sensorML.xsd";

    static final HashMap<String, String> FIXED_SUGGESTED_PREFIXES;

    public static final XmlOptions DEFAULT_OPTIONS = new XmlOptions();

    static {
        FIXED_SUGGESTED_PREFIXES = new HashMap<String, String>();
        FIXED_SUGGESTED_PREFIXES.put(SOS_1_0_NAMESPACE_URI, SOS_NAMESPACE_PREFIX);
        FIXED_SUGGESTED_PREFIXES.put(SWE_1_0_1_NAMESPACE_URI, SWE_NAMESPACE_PREFIX);
        FIXED_SUGGESTED_PREFIXES.put(OM_1_0_NAMESPACE_URI, OM_NAMESPACE_PREFIX);
        FIXED_SUGGESTED_PREFIXES.put(SA_1_0_NAMESPACE_URI, SA_NAMESPACE_PREFIX);
        FIXED_SUGGESTED_PREFIXES.put(GML_NAMESPACE_URI, GML_NAMESPACE_PREFIX);
        FIXED_SUGGESTED_PREFIXES.put(OWS_1_1_NAMESPACE_URI, OWS_NAMESPACE_PREFIX);
        FIXED_SUGGESTED_PREFIXES.put(SML_1_0_1_NAMESPACE_URI, SML_NAMESPACE_PREFIX);
        FIXED_SUGGESTED_PREFIXES.put(XMLConstants.W3C_XML_SCHEMA_INSTANCE_NS_URI, XSI_NAMESPACE_PREFIX);

        DEFAULT_OPTIONS.setSaveNamespacesFirst();
        DEFAULT_OPTIONS.setSavePrettyPrint();
        DEFAULT_OPTIONS.setSaveSuggestedPrefixes(FIXED_SUGGESTED_PREFIXES);
        DEFAULT_OPTIONS.setSaveAggressiveNamespaces();
        DEFAULT_OPTIONS.setSavePrettyPrintIndent(DEFAULT_INDENT);
    }

    /**
     * Pretty-prints xml, supplied as a string.
     * 
     * Kudos to http://stackoverflow.com/questions/139076/how-to-pretty-print-xml-from-java
     * <p/>
     * eg. <code>
     * String formattedXml = new XmlFormatter().format("<tag><nested>hello</nested></tag>");
     * </code>
     */
    public static String format(String unformattedXml) {
        try {
            final Document document = parseXmlFile(unformattedXml);

            OutputFormat format = new OutputFormat(document);
            format.setLineWidth(65);
            format.setIndenting(true);
            format.setIndent(2);
            Writer out = new StringWriter();
            XMLSerializer serializer = new XMLSerializer(out, format);
            serializer.serialize(document);

            return out.toString();
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 
     * @param node
     * @return
     */
    public static String xmlToString(Node node) {
        try {
            Source source = new DOMSource(node);
            StringWriter stringWriter = new StringWriter();
            Result result = new StreamResult(stringWriter);
            TransformerFactory factory = TransformerFactory.newInstance();
            Transformer transformer = factory.newTransformer();
            transformer.transform(source, result);

            return format(stringWriter.getBuffer().toString());
        }
        catch (TransformerConfigurationException e) {
            e.printStackTrace();
        }
        catch (TransformerException e) {
            e.printStackTrace();
        }
        return "ERROR transforming given Node";
    }

    /**
     * 
     * Returns {@link XmlObject#xmlText()}, but only a maximum of {@link XmlTools#MAX_INSPECT_STRING_LENGTH}
     * characters.
     * 
     * @param xmlObj
     * @return
     */
    public static String inspect(XmlObject xmlObj) {
        String s = xmlObj.xmlText();

        // remove comment at start
        if (s.startsWith("<!--")) {
            int commentEnd = s.indexOf("-->") + 3;
            s = s.substring(commentEnd, s.length() - commentEnd);
        }

        if (s.length() <= MAX_INSPECT_STRING_LENGTH)
            return s;
        return s.substring(0, MAX_INSPECT_STRING_LENGTH) + "... ( " + (s.length() - MAX_INSPECT_STRING_LENGTH)
                + " more characters)";
    }

    public static String inspect(Document document) {
        return document.getNodeValue();
    }

    /**
     * http://xmlbeans.apache.org/docs/2.0.0/guide/conValidationWithXmlBeans.html
     * 
     * @param xml
     * @return
     */
    public static String validateAndIterateErrors(XmlObject xml) {
        ArrayList<XmlError> validationErrors = new ArrayList<XmlError>();
        XmlOptions validationOptions = new XmlOptions();
        validationOptions.setErrorListener(validationErrors);

        boolean isValid = xml.validate(validationOptions);

        StringBuilder sb = new StringBuilder();
        if ( !isValid) {
            sb.append("XmlObject of class <");
            sb.append(xml.getClass().getSimpleName());
            sb.append(" /> is NOT valid! The validation errors are:");

            Iterator<XmlError> iter = validationErrors.iterator();
            while (iter.hasNext()) {
                sb.append("\n\t");
                sb.append(iter.next().toString());
            }
        }
        else {
            sb.append("XmlObject ");
            sb.append(xml.getClass().getSimpleName());
            sb.append(" (");
            sb.append(xml.xmlText().substring(0, 40));
            sb.append("...) is valid!");
        }

        return sb.toString();
    }

    public static Map<String, String> getFixedSuggestedPrefixes() {
        return FIXED_SUGGESTED_PREFIXES;
    }

    private static Document parseXmlFile(String in) {
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(in));
            return db.parse(is);
        }
        catch (ParserConfigurationException e) {
            throw new RuntimeException(e);
        }
        catch (SAXException e) {
            throw new RuntimeException(e);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

}